package geometry.part5;

import geometry.common.Demo;
import geometry.common.DemoFrame;
import geometry.common.DemoPanel;
import geometry.part2.LineSegment;
import geometry.part3.Intersection;
import geometry.part4.Event;
import geometry.part4.SweepLineBasedComparator;

import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;

public class PlaneSweepDemo {
    public static void main(String[] args) {
        Demo.launch(IntersectionFrame.class, "平面走査法による線分の交差検出");
    }

    public static class IntersectionFrame extends DemoFrame {
        private static final long serialVersionUID = 1L;

        enum Mode {
            MOVE, ADD, DELETE
        }

        enum MoveTarget {
            WHOLE, POINT1, POINT2;
        }

        List<LineSegment> segments;
        SweepLineBasedComparator sweepComparator;
        TreeSet<Event> eventQueue;
        TreeSet<LineSegment> status;
        Set<Intersection> result;
        Set<Intersection> highlightResult;

        JToggleButton addButton;
        JToggleButton moveButton;
        JToggleButton deleteButton;
        Event event;
        Mode mode;
        LineSegment highlightSegment;
        Point pressPoint;
        LineSegment baseSegment;
        LineSegment workingSegment;
        MoveTarget workingTarget;

        @Override
        protected void initialize() {
            segments = new ArrayList<LineSegment>();
            segments.add(new LineSegment(200, 100, 400, 450));
            segments.add(new LineSegment(350, 60, 230, 360));
            segments.add(new LineSegment(150, 220, 500, 120));
            segments.add(new LineSegment(130, 320, 280, 500));
            segments.add(new LineSegment(280, 340, 170, 470));

            sweepComparator = new SweepLineBasedComparator();
            eventQueue = new TreeSet<Event>();
            status = new TreeSet<LineSegment>(sweepComparator);
            result = new HashSet<Intersection>();
            highlightResult = new HashSet<Intersection>();

            moveButton = new JToggleButton("線分の移動");
            addButton = new JToggleButton("線分の追加");
            deleteButton = new JToggleButton("線分の削除");
            ButtonGroup group = new ButtonGroup();
            group.add(moveButton);
            group.add(addButton);
            group.add(deleteButton);
            group.setSelected(moveButton.getModel(), true);

            JButton nextButton = new JButton("次のイベント");
            JButton resetButton = new JButton("走査線のリセット");
            nextButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    next();
                }
            });
            resetButton.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    reset();
                }
            });

            JToolBar toolBar = new JToolBar();
            toolBar.setFloatable(false);
            toolBar.add(moveButton);
            toolBar.add(addButton);
            toolBar.add(deleteButton);
            toolBar.addSeparator();
            toolBar.add(nextButton);
            toolBar.add(resetButton);

            final IntersectionPanel panel = new IntersectionPanel();
            panel.addMouseListener(new MouseAdapter() {
                @Override
                public void mousePressed(MouseEvent e) {
                    Point p = e.getPoint();
                    pressPoint = p;
                    switch (getMode()) {
                    case MOVE:
                        workingSegment = getSegmentAt(p);
                        if (workingSegment != null) {
                            if (p
                                    .distance(workingSegment.x1,
                                            workingSegment.y1) <= 8) {
                                workingTarget = MoveTarget.POINT1;
                            } else if (p.distance(workingSegment.x2,
                                    workingSegment.y2) <= 8) {
                                workingTarget = MoveTarget.POINT2;
                            } else {
                                workingTarget = MoveTarget.WHOLE;
                                baseSegment = new LineSegment(
                                        workingSegment.x1, workingSegment.y1,
                                        workingSegment.x2, workingSegment.y2);
                            }
                        }
                        break;
                    case ADD:
                        double x = e.getX();
                        double y = e.getY();
                        workingSegment = new LineSegment(x, y, x, y);
                        break;
                    case DELETE:
                        LineSegment s = getSegmentAt(p);
                        if (s != null) {
                            segments.remove(s);
                            reset();
                        }
                        break;
                    }
                }

                @Override
                public void mouseReleased(MouseEvent e) {
                    switch (getMode()) {
                    case ADD:
                        if (workingSegment.x1 != workingSegment.x2
                                || workingSegment.y1 != workingSegment.y2) {
                            segments.add(workingSegment);
                            reset();
                        }
                        break;
                    }
                    workingSegment = null;
                    repaint();
                }
            });
            panel.addMouseMotionListener(new MouseMotionAdapter() {
                @Override
                public void mouseMoved(MouseEvent e) {
                    LineSegment s = getSegmentAt(e.getPoint());
                    if (s != highlightSegment) {
                        highlightSegment = s;
                        repaint();
                    }
                }

                @Override
                public void mouseDragged(MouseEvent e) {
                    int x = e.getX();
                    int y = e.getY();
                    switch (getMode()) {
                    case MOVE:
                        if (workingSegment != null) {
                            switch (workingTarget) {
                            case WHOLE:
                                int dx = x - pressPoint.x;
                                int dy = y - pressPoint.y;
                                workingSegment.x1 = baseSegment.x1 + dx;
                                workingSegment.y1 = baseSegment.y1 + dy;
                                workingSegment.x2 = baseSegment.x2 + dx;
                                workingSegment.y2 = baseSegment.y2 + dy;
                                break;
                            case POINT1:
                                workingSegment.x1 = x;
                                workingSegment.y1 = y;
                                break;
                            case POINT2:
                                workingSegment.x2 = x;
                                workingSegment.y2 = y;
                                break;
                            }
                            reset();
                        }
                        break;
                    case ADD:
                        workingSegment.x2 = x;
                        workingSegment.y2 = y;
                        repaint();
                        break;
                    }
                }
            });

            Container c = getContentPane();
            c.setLayout(new BorderLayout());
            c.add(toolBar, BorderLayout.NORTH);
            c.add(panel, BorderLayout.CENTER);

            reset();
        }

        private Mode getMode() {
            if (addButton.isSelected()) {
                return Mode.ADD;
            } else if (moveButton.isSelected()) {
                return Mode.MOVE;
            } else {
                return Mode.DELETE;
            }
        }

        private LineSegment getSegmentAt(Point2D p) {
            for (LineSegment s : segments) {
                if (Line2D
                        .ptSegDist(s.x1, s.y1, s.x2, s.y2, p.getX(), p.getY()) <= 8) {
                    return s;
                }
            }
            return null;
        }

        private void next() {
            if (event == null) {
                event = eventQueue.first();
            } else {
                event = eventQueue.higher(event);
            }
            if (event != null) {
                highlightResult.clear();
                double sweepY = event.y;
                switch (event.type) {
                case SEGMENT_START:
                    sweepComparator.setY(sweepY);
                    LineSegment newSegment = event.segment1;
                    status.add(newSegment);
                    LineSegment left = status.lower(newSegment);
                    LineSegment right = status.higher(newSegment);
                    checkIntersection(left, newSegment, sweepY);
                    checkIntersection(newSegment, right, sweepY);
                    break;
                case INTERSECTION:
                    left = event.segment1;
                    right = event.segment2;
                    LineSegment moreLeft = status.lower(left);
                    LineSegment moreRight = status.higher(right);
                    status.remove(left);
                    status.remove(right);
                    sweepComparator.setY(sweepY);
                    if (sweepComparator.compare(left, right) < 0) {
                        sweepComparator.setY(sweepY + 0.001);
                    }
                    status.add(left);
                    status.add(right);
                    checkIntersection(moreLeft, right, sweepY);
                    checkIntersection(left, moreRight, sweepY);
                    break;
                case SEGMENT_END:
                    LineSegment endSegment = event.segment1;
                    left = status.lower(endSegment);
                    right = status.higher(endSegment);
                    checkIntersection(left, right, sweepY);
                    status.remove(endSegment);
                    sweepComparator.setY(sweepY);
                    break;
                }
                repaint();
            } else {
                reset();
            }
        }

        private void reset() {
            event = null;
            eventQueue.clear();
            for (LineSegment s : segments) {
                if (s.y1 < s.y2 || (s.y1 == s.y2 && s.x1 < s.x2)) {
                    eventQueue.add(new Event(Event.Type.SEGMENT_START, s.x1,
                            s.y1, s, null));
                    eventQueue.add(new Event(Event.Type.SEGMENT_END, s.x2,
                            s.y2, s, null));
                } else {
                    eventQueue.add(new Event(Event.Type.SEGMENT_START, s.x2,
                            s.y2, s, null));
                    eventQueue.add(new Event(Event.Type.SEGMENT_END, s.x1,
                            s.y1, s, null));
                }
            }
            status.clear();
            result.clear();
            highlightResult.clear();
            repaint();
        }

        private void checkIntersection(LineSegment left, LineSegment right,
                double sweepY) {
            if (left == null || right == null) {
                return;
            }
            Point2D p = left.getIntersectionPoint(right);
            if (p != null && p.getY() >= sweepY) {
                eventQueue.add(new Event(Event.Type.INTERSECTION, p.getX(), p
                        .getY(), left, right));
                Intersection i = new Intersection(left, right);
                if (!result.contains(i)) {
                    result.add(i);
                    highlightResult.add(i);
                }
            }
        }

        private class IntersectionPanel extends DemoPanel {
            private static final long serialVersionUID = 1L;

            @Override
            protected void doPaint(Graphics2D g) {
                g.setStroke(new BasicStroke(1));
                for (LineSegment s : segments) {
                    if (s == highlightSegment) {
                        drawSegment(g, s, Color.BLACK, 3);
                    } else {
                        drawSegment(g, s, Color.BLACK, 1);
                    }
                }
                if (workingSegment != null) {
                    drawSegment(g, workingSegment, Color.BLACK, 3);
                }

                if (event != null) {
                    LineSegment s1 = event.segment1;
                    LineSegment s2 = event.segment2;
                    switch (event.type) {
                    case SEGMENT_START:
                        LineSegment l = status.lower(s1);
                        LineSegment r = status.higher(s1);
                        drawSegment(g, s1, Color.RED, 3);
                        drawSegment(g, l, Color.BLUE, 3);
                        drawSegment(g, r, Color.BLUE, 3);
                        break;
                    case INTERSECTION:
                        if (sweepComparator.compare(s1, s2) < 0) {
                            l = status.lower(s1);
                            r = status.higher(s2);
                        } else {
                            l = status.lower(s2);
                            r = status.higher(s1);
                        }
                        drawSegment(g, s1, Color.RED, 3);
                        drawSegment(g, s2, Color.RED, 3);
                        drawSegment(g, l, Color.BLUE, 3);
                        drawSegment(g, r, Color.BLUE, 3);
                        break;
                    case SEGMENT_END:
                        l = status.lower(s1);
                        r = status.higher(s1);
                        drawSegment(g, s1, Color.RED, 3);
                        drawSegment(g, l, Color.BLUE, 3);
                        drawSegment(g, r, Color.BLUE, 3);
                        break;
                    }

                    for (Intersection i : result) {
                        Point2D p = i.getIntersectionPoint();
                        g.setPaint(highlightResult.contains(i) ? Color.BLUE
                                : Color.DARK_GRAY);
                        g.fill(new Ellipse2D.Double(p.getX() - 8, p.getY() - 8,
                                16, 16));
                    }

                    double x = event.x;
                    double y = event.y;
                    g.setPaint(Color.BLACK);
                    g.setStroke(new BasicStroke(2f));
                    g.draw(new Line2D.Double(x - 8, y - 8, x + 8, y + 8));
                    g.draw(new Line2D.Double(x + 8, y - 8, x - 8, y + 8));

                    g
                            .setStroke(new BasicStroke(1,
                                    BasicStroke.CAP_SQUARE,
                                    BasicStroke.JOIN_BEVEL, 1f, new float[] {
                                            2, 3 }, 1));
                    g.draw(new Line2D.Double(0, y, getWidth(), y));
                }
            }

            private void drawSegment(Graphics2D g, LineSegment s, Color color,
                    float width) {
                if (s == null) {
                    return;
                }
                g.setPaint(color);
                g.setStroke(new BasicStroke(width));
                g.draw(new Line2D.Double(s.x1, s.y1, s.x2, s.y2));
                g.fill(new Ellipse2D.Double(s.x1 - 4, s.y1 - 4, 8, 8));
                g.fill(new Ellipse2D.Double(s.x2 - 4, s.y2 - 4, 8, 8));
            }
        }
    }
}
